 /* host_config.c - host configuration file
 *
 * Copyright (C) 2021 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <uci.h>

#include "debug.h"
#include "host_config.h"
#include "topologyd.h"
#include "host.h"

struct uci_section *config_get_section(struct uci_context *ctx,
		struct uci_package *pkg, const char *type, const char *key,
		const char *value)
{
	struct uci_element *e;
	struct uci_section *section;

	dbg("Inside %s %d\n", __func__, __LINE__);

	/* get the wet iface section */
	uci_foreach_element(&pkg->sections, e) {
		const char *c_value;

		section = uci_to_section(e);
		if (strcmp(section->type, type))
			continue;

		c_value = uci_lookup_option_string(ctx, section, key);
		dbg("Inside %s %d\n %s-------%s", __func__, __LINE__, c_value, value);
		if (c_value && !strcmp(c_value, value))
			return section;
	}

	return NULL;
}

int config_del_section(const char *config, const char *type, const char *value, bool persistent)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_section *section;
	struct uci_ptr ptr = {0};
	int rv = -1;

	dbg("Inside %s %d the value is %s\n", __func__, __LINE__, value);

	ctx = uci_alloc_context();
	if (!ctx)
		goto out;

	if (persistent == 0) {
		uci_set_confdir(ctx, "/var/state");
	}

	if (uci_load(ctx, config, &pkg) != UCI_OK) {
		dbg("config file not found!\n");
		goto out_uci;
	}

	section = config_get_section(ctx, pkg, type, "macaddr", value);
	if (!section)
		goto out_pkg;

	ptr.p = pkg;
	ptr.s = section;

	uci_delete(ctx, &ptr);
	uci_commit(ctx, &pkg, false);
out_pkg:
	uci_unload(ctx, pkg);
out_uci:
	uci_free_context(ctx);
out:
	return rv;
}


struct uci_section *config_add_section(struct uci_context *ctx,
		struct uci_package *pkg, const char *config, const char *type,
		char *name, const char *key, const char *value)
{
	struct uci_section *section = NULL;
	struct uci_ptr ptr = {0};
	int rv = -1;
	char name_section[50] = { 0 };
	static const char s[2] = ":";
	char *token;
	char mac_name[18] = { 0 };


	dbg("Inside %s %d\n", __func__, __LINE__);
	strncpy(mac_name, value, 18);
	/* get the first token */
	token = strtok(mac_name, s);

	/* walk through other tokens */
	while (token != NULL) {
		snprintf(name_section + strlen(name_section),
			sizeof(name_section), "%s", token);
		token = strtok(NULL, s);
	}

	section = config_get_section(ctx, pkg, type, key, value);
	if (!section) {
		rv = uci_add_section(ctx, pkg, type, &section);
		if (rv)
			goto out_pkg;

		ptr.p = pkg;
		ptr.s = section;
		ptr.value = name_section;
		uci_rename(ctx, &ptr);
		rv = uci_save(ctx, pkg);
		if (rv)
			goto out_pkg;
	}

	ptr.value = value;
	ptr.package = config;
	ptr.section = section->e.name;
	ptr.option = key;
	ptr.target = UCI_TYPE_OPTION;

	uci_lookup_ptr(ctx, &ptr, NULL, false);
	uci_set(ctx, &ptr);
	uci_save(ctx, ptr.p);

out_pkg:
	return section;
}


int config_add_default_host_mac(const char *config, const char *type,
		const char *macaddr, bool persistent)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_section *section;

	dbg("Inside %s %d\n", __func__, __LINE__);

	ctx = uci_alloc_context();
	if (!ctx)
		goto out;

	if (persistent == 0) {
		uci_set_confdir(ctx, "/var/state");
	}

	if (uci_load(ctx, config, &pkg) != UCI_OK) {
		err("config file 'host' not found!\n");
		goto out_uci;
	}

	section = config_add_section(ctx, pkg, config, type, "mac_addr", "macaddr", macaddr);
	if (!section)
		return -1;

	uci_commit(ctx, &pkg, false);
	uci_unload(ctx, pkg);
out_uci:
	uci_free_context(ctx);
out:
	return 0;
}

bool config_set_host_option(char *package_name,
		char *section_type, char *search_key, char *search_val,
		char *option, char *value, bool persistent)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!package_name || !search_val || !option || !value)
		return false;

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	if (persistent == 0) {
		uci_set_confdir(ctx, "/var/state");
	}

	if (uci_load(ctx, package_name, &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, section_type)) {
			struct uci_option *opt = uci_lookup_option(ctx, s,
							search_key);

			if (!opt || opt->type != UCI_TYPE_STRING)
				continue;
			if (strcmp(opt->v.string, search_val) == 0) {
				struct uci_ptr ptr = {0};

				ptr.value = value;
				ptr.package = package_name;
				ptr.section = s->e.name;
				ptr.option = option;
				ptr.target = UCI_TYPE_OPTION;
				if (uci_lookup_ptr(ctx, &ptr, NULL, false) ||
						!UCI_LOOKUP_COMPLETE)
					break;
				if (uci_set(ctx, &ptr) == UCI_OK)
					uci_save(ctx, ptr.p);
				break;
			}
		}
	}
	uci_commit(ctx, &pkg, false);
	uci_unload(ctx, pkg);
	uci_free_context(ctx);
	return false;
}

bool config_init_host_data(char *package_name, char *section, struct topologyd_private *priv)
{

	bool ret;
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	char mac_address[18];
	uint8_t mac[6];
	char intf_type[20];

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!package_name)
		return false;

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	if (priv->host_global_cfg.persistent == 0) {
		uci_set_confdir(ctx, "/var/state");
	}

	if (uci_load(ctx, package_name, &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	ret = false;
	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, section)) {
			struct host_node *p = NULL;
			struct uci_option *opt = uci_lookup_option(ctx, s,
							"macaddr");
			if (!opt || opt->type != UCI_TYPE_STRING)
				continue;
			strncpy(mac_address, opt->v.string, 18);
			hwaddr_aton(mac_address, mac);

			opt = uci_lookup_option(ctx, s, "interface_type");
			if (!opt || opt->type != UCI_TYPE_STRING)
				continue;
			strncpy(intf_type, opt->v.string, 20);
			if (!strcmp(intf_type, "eth")) {
				dbg("%s IS AN ETHERNET CLIENT ACCORDING TO CONFIG\n", mac_address);
				p = host_topo_node_add(priv, NULL, mac, 0);
			} else {
				dbg("%s IS AN WIFI CLIENT ACCORDING TO CONFIG\n", mac_address);
				p = host_topo_node_add(priv, NULL, mac, 1);
			}
			if (p) {
				if (p->intf_type == HOST_TYPE_WIFI) {
					// We are changing all to zero in init so that it can be correctly changed after the
					// topology extended response message
					host_change_active_state(priv, p, 0);
				}
			}
		}
	}
	uci_unload(ctx, pkg);
	uci_free_context(ctx);
	return ret;
}

struct uci_package *config_init_package(struct uci_context **ctx, const char *config)
{
	struct uci_package *p = NULL;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (*ctx == NULL) {
		*ctx = uci_alloc_context();
	} else {
		p = uci_lookup_package(*ctx, config);
		if (p)
			uci_unload(*ctx, p);
	}

	if (uci_load(*ctx, config, &p))
		return NULL;

	return p;
}

void config_free_uci_context(struct uci_context **ctx)
{
	dbg("Inside %s %d\n", __func__, __LINE__);
	if (*ctx)
		uci_free_context(*ctx);
	*ctx = NULL;
}

static int config_get_section_mapagent(bool *enable,
		struct uci_section *s)
{
	dbg("Inside %s %d\n", __func__, __LINE__);

	enum {
		AGENT_ATTR_ENABLED,
		AGENT_ATTR_NUM
	};
	const struct uci_parse_option opts[] = {
		{ .name = "enabled", .type = UCI_TYPE_STRING },
	};
	struct uci_option *tb[1];

	uci_parse_section(s, opts, 1, tb);

	if (tb[AGENT_ATTR_ENABLED]) {
		const char *val = tb[AGENT_ATTR_ENABLED]->v.string;

		*enable = atoi(val) == 1 ? true : false;
	}
	return 0;
}

int host_config_mapagent_get(bool *enable)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	dbg("Inside %s %d\n", __func__, __LINE__);
	ctx = uci_alloc_context();
	if (!ctx)
		return -1;

	if (uci_load(ctx, MAPAGENT_CONFIG, &pkg)) {
		uci_free_context(ctx);
		return -1;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, "agent"))
			config_get_section_mapagent(enable, s);
	}

	uci_unload(ctx, pkg);
	uci_free_context(ctx);

	return 0;
}

void host_config_dump(struct host_config *cfg)
{
	int i = 0;

	dbg("Host config ---------\n");
	dbg("Ageing Timer                  : %d\n", cfg->age_timer);
	dbg("Reboot Persistent             : %d\n", cfg->persistent);
	for (i = 0; i < cfg->num_ifname; i++)
		dbg("ifname             : %s\n", cfg->ifname[i]);
	dbg("---------------------------\n");
}

static int config_get_global_section_host(struct host_config *cfg,
		struct uci_section *s)
{
	enum {
		HOST_ATTR_AGEING_TIMER,
		HOST_ATTR_REBOOT_PERSIST,
		HOST_ATTR_IFNAME,
		NUM_HOST_ATTRS,
	};
	const struct uci_parse_option opts[] = {
		{ .name = "ageing_timer", .type = UCI_TYPE_STRING },
		{ .name = "reboot_persistent", .type = UCI_TYPE_STRING },
		{ .name = "ifname", .type = UCI_TYPE_LIST }
	};
	struct uci_option *tb[NUM_HOST_ATTRS];
	uci_parse_section(s, opts, NUM_HOST_ATTRS, tb);

	if (tb[HOST_ATTR_AGEING_TIMER]) {
		const char *val = tb[HOST_ATTR_AGEING_TIMER]->v.string;

		int v = atoi(val);

		if (v > 0)
			cfg->age_timer = v;
	}

	if (tb[HOST_ATTR_REBOOT_PERSIST]) {
		const char *val = tb[HOST_ATTR_REBOOT_PERSIST]->v.string;

		cfg->persistent = atoi(val) == 1 ? true : false;
	}

	if (tb[HOST_ATTR_IFNAME]) {
		struct uci_element *x;
		int i = 0;

		uci_foreach_element(&tb[HOST_ATTR_IFNAME]->v.list, x) {
			if (cfg->num_ifname >= MAX_IFNAME) {
				dbg("at most %d ifname\n", MAX_IFNAME);
				break;
			}
			strncpy((char *)(cfg->ifname[i]), x->name, 15);
			i++;
			cfg->num_ifname = i;
		}
	}

	return 0;
}

static bool config_get_ifname(struct uci_section *s, char *ifname, int *num_ifname)
{
	bool ret = false;
	enum {
		HOST_ATTR_IFNAME,
		NUM_HOST_ATTRS,
	};
	const struct uci_parse_option opts[] = {
		{ .name = "ifname", .type = UCI_TYPE_LIST }
	};
	struct uci_option *tb[NUM_HOST_ATTRS];

	dbg("Inside %s %d\n", __func__, __LINE__);

	uci_parse_section(s, opts, NUM_HOST_ATTRS, tb);

	if (tb[HOST_ATTR_IFNAME]) {
		struct uci_element *x;
		int i = 0;

		uci_foreach_element(&tb[HOST_ATTR_IFNAME]->v.list, x) {
			if (*num_ifname >= MAX_IFNAME) {
				dbg("more than max iface %d ifname\n", MAX_IFNAME);
				break;
			}
			strncpy(&(ifname[i*16]), x->name, 15);
			i++;
			*num_ifname = i;
			ret = true;
		}
	}
	dbg("Inside %s %d num_iface = %d \n", __func__, __LINE__, *num_ifname);

	return ret;
}

int host_config_global(struct host_config *cfg)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	ctx = uci_alloc_context();
	if (!ctx)
		return -1;

	if (uci_load(ctx, HOST_CONFIG, &pkg)) {
		uci_free_context(ctx);
		return -1;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, "global")) {
			config_get_global_section_host(cfg, s);
		}
	}

	host_config_dump(cfg);

	uci_unload(ctx, pkg);
	uci_free_context(ctx);

	return 0;
}

bool config_get_al_ifname(char *ifname, int *num_iface)
{

	bool ret = false;
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	//char intf_type[20];

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!ifname)
		return false;

	ctx = uci_alloc_context();
	if (!ctx)
		return false;


	if (uci_load(ctx, "ieee1905", &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, "al-iface"))
			ret = config_get_ifname(s, ifname, num_iface);
	}
	uci_unload(ctx, pkg);
	uci_free_context(ctx);
	return ret;
}
